use POSIX qw(strftime tmpnam);
use File::Spec::Functions qw(rel2abs file_name_is_absolute);
use fidoaddr;
use misc;

{
  package config;
  my $mfile; # main config-file
  my $cfile; # current config-file
  my $cline; # current config-line
  my $afile; # config-file with last found areas record
1;
}

# keywords hash, value is bit field
# bit 0: mandatory keyword
# bit 1: allowed multiple
# bit 4: check path
# bit 5: add base_path if not absolute
%keywords = ('address'=>1, 'aka'=>2, 'inbound'=>0x33, 'outbound'=>0x32,
             't-mailbox'=>0x30, 't-maillongbox'=>0x30, 'thebrake!box'=>0x30,
             'mailin'=>0x30, 'mailout'=>0x30, 'echoout'=>0, 
             'badtics'=>0x31, 'dupetics'=>0x30, 'proctics'=>0x30, 'queue'=>0x31, 
             'logfile'=>0x20, 'dupefile'=>0x20, 'traffile'=>0x20, 'repfile'=>0x20, 
             'lockfile'=>0x20, 'locktime'=>0, 'magicfile'=>2, 'magicpass'=>0,
             'fecholist'=>2,
             'group'=>2, 'crc32'=>0x20, 'include'=>0x22, 'attachmode'=>0, 'filemode'=>0,
             'loglevel'=>0, 'conlevel'=>0, 'robot'=>1, 'sysop'=>0, 'origin'=>0,
             'splitmsg'=>0, 'filesbbs'=>0, 'filekeepdays'=>0, 'dupekeepdays'=>0,
             'feedunlinks'=>0, 'lastunlinks'=>0, 'guesslongdesc'=>0,
             'node'=>2, 'area'=>2);
# node keywords hash, value is bit field
%nodekeys = ('useaka'=>0, 'sysop'=>0, 'robot'=>0, 'groups'=>0, 'box'=>0,
             'flag'=>0, 'options'=>0, 'arealist'=>0, 'autoconnect'=>2);
# area keywords hash, value is bit field
%areakeys = ('links'=>2, 'description'=>0, 'group'=>0, 'exclude'=>0,
             'restrict'=>0, 'dupecheck'=>0, 'nocheck'=>0, 'magic'=>2, 'file'=>2,
             'onrecv'=>0, 'onhatch'=>0, 'oncreate'=>0);
# --------------------------------------------------------------------
# print error in config-file and exit
sub cfg_error
{
  print "(".$config::cfile.":".$config::cline.") $_[0]"; exit;
}
# --------------------------------------------------------------------
# function parse_flag(string $s)
#    parses flags from $s, return value
sub parse_flag
{
  my $s = $_[0];
  if ($s =~ /^(h|hld|hold)$/i) { return 'h'; }
  elsif ($s =~ /^(c|cra|crash)$/i) { return 'c'; }
  elsif ($s =~ /^(d|dir|direct)$/i) { return 'd'; }
  elsif ($s =~ /^(i|imm)$/i) { return 'i'; }
  else { return undef; }
}
# --------------------------------------------------------------------
# function init_config()
#    set default values
sub init_config
{
  %config = ();
  $config{'conlevel'} = $config{'loglevel'} = 2;
  $config{'attachmode'} = 'single';
  $config{'filemode'} = 'long';
  $config{'crc_seek'} = 'missing';
  $config{'logfile'}  = 'ffix.log';
  $config{'dupefile'} = 'ffix.dup';
  $config{'traffile'} = 'ffix.trf';
  $config{'repfile'}  = 'ffix.rep';
  $config{'filekeepdays'} = 0;
  $config{'dupekeepdays'} = 30;
}
# --------------------------------------------------------------------
# function read_config(string $name)
#    read config file $name and set variables %config, %nodes, %areas
sub read_config
{
  # vars to track duplicate keywords
  my %was, %was_n, %was_k;
  # vars to keep state
  my $in_area = 0; my $in_node = 0;
  # open file
  my $file = $_[0];
  $config::cfile = $file; $config::mfile = $file unless $config::mfile ne '';
  if (!open F, $file) {
    say 2, "Error reading config file $file"; return 0;
  }
  say 5, "Reading config file $file";
  my @a = <F>;
  close F;
  for (my $i = 1; $i <= @a; $i++) {
    my $cmd; 
    my $type = -1;
    $config::cline = $i;
    $_ = $a[$i-1]; chomp $_;
    # strip whitespaces
    s/^\s+//; s/\s+$//;
    # ignore comments
    if ( m!^(#|;|//)! ) { next; }
    # ignore empty lines
    if ( m!^\s*$! ) { next; }
    # get keyword as first word
    ($cmd) = /^(\S+)/; $cmd = lc $cmd;
    if ( defined $keywords{$cmd} ) { $type = $keywords{$cmd}; }
    if ( defined $nodekeys{$cmd} ) { 
      if ($in_node) { $type = $nodekeys{$cmd}; }
      elsif ($type < 0) { cfg_error("Keyword '$cmd' allowed only in 'node' section\n"); }
    }
    if ( defined $areakeys{$cmd} ) { 
      if ($in_area) { $type = $areakeys{$cmd}; }
      elsif ($type < 0) { cfg_error("Keyword '$cmd' allowed only in 'area' section\n"); }
    }
    if ( $type < 0 ) {
      cfg_error("Unknown keyword '$cmd'\n");
    }
    # check duplicate keywords
    my $_was = 0;
    if ($in_node) { $_was = $was_n{$cmd}; } elsif ($in_area) { $_was = $was_a{$cmd}; } else { $_was = $was{$cmd}; }
    if ( !($type & 2) && ($_was > 0) ) {
      cfg_error("Keyword '$cmd' already defined on line $_was\n");
    }
    # parse commands
    study $_;
    # Address <main fido address 5d>
    if ($cmd eq 'address') {
      if (!/^\S+\s+(.*)$/) { cfg_error("Syntax error\n"); }
      my @addr = s2addr($1);
      if ( !defined $addr[0] ) { cfg_error("Bad address $1\n"); }
      $config{'address'} = \@addr;
      @node_me = @addr[0..3];
    }
    # Aka <aka 5d> <mask to match this aka>
    if ($cmd eq 'aka') {
      if (!/^\S+\s+(\S+)\s+(\S+)$/) { cfg_error("Syntax error\n"); }
      my @addr1 = s2addr($1); my @addr2 = s2addr($2);
      if ( !defined $addr1[0] ) { cfg_error("Bad address $1\n"); }
      if ( !defined $addr2[0] ) { cfg_error("Bad address $2\n"); }
      #push @{$config{'aka'}}, { addr2s(@addr1)=>\@addr2 };
      $config{'aka'}{addr2s(@addr1)} = \@addr2;
    }
    # Inbound <path>
    if ($cmd eq 'inbound') { /^\S+\s+(.*)$/; push @{$config{'inbound'}}, parse_quote2($1, '"'); }
    # Outbound <path> [<mask>]
    if ($cmd eq 'outbound') { 
      if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(?:\s+(.*?))?$/) { cfg_error("Syntax error"); }
      my $path = parse_quote($1, '"').$2;
      if ( defined $3 ) {
        my @addr = s2addr($3);
        if ( !defined $addr[0] ) { cfg_error("Bad address $3\n"); }
        push @{$config{'outb_dir'}}, [$path, @addr];
      }
      else { $config{'outbound'} = $path; }
#      /^\S+\s+(.*)$/; $config{'outbound'} = parse_quote2($1, '"'); 
    }
    # AttachMode <link|symlink|single>
    if ($cmd eq 'attachmode') { 
      if ( !/^\S+\s+(symlink|link|single)$/i ) { cfg_error("Attach mode must be 'symlink', 'link' or 'single'"); }
      $config{$cmd} = lc $1; 
    }
    # T-MailBox     <path> [default] [notic]
    # T-MailLongBox <path> [default] [notic]
    # TheBrake!Box  <path> [default] [notic]
    if ($cmd eq 't-mailbox' || $cmd eq 't-maillongbox' || $cmd eq 'thebrake!box') {
      if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(.*?)$/) { cfg_error("Syntax error"); }
      my $path = parse_quote($1, '"').$2; my $opts = $3;
      $config{$cmd} = $path;
      if ($opts =~ /default/i) { 
        if (defined $config{'default_box'}) { cfg_error("Default filebox type already defined"); }
        $config{'default_box'} = 1 if $cmd eq 't-mailbox';
        $config{'default_box'} = 2 if $cmd eq 't-maillongbox';
        $config{'default_box'} = 3 if $cmd eq 'thebrake!box';
      }
    }
    # FileMode <long|short|both>
    if ($cmd eq 'filemode') { 
      if ( !/^\S+\s+(short|long|both)$/i ) { cfg_error("File mode must be 'short', 'long' or 'both'"); }
      $config{$cmd} = lc $1; 
    }
    # MailIn <path>
    if ($cmd eq 'mailin') { /^\S+\s+(.*)$/; $config{'mailin'} = parse_quote2($1, '"'); }
    # MailOut <path> [<msg|pkt|pkt[2[+|o]]>]
    if ($cmd eq 'mailout') {
      /^\S+\s+(.*)$/; 
      my $s = $1;
      #if ($s =~ /\s+pkt$/i) { $config{'mailout_pkt'} = 1; }
      $s =~ s/\s+(msg|pkt(?:2(?:[o+])?)?)$//i;
      if ( defined $1 ) { 
        if ( lc($1) eq 'pkt' || lc($1) eq 'pkt2' ) { $config{'mailout_pkt'} = 21; }
        elsif ( lc($1) eq 'pkt2o' ) { $config{'mailout_pkt'} = 20; }
        elsif ( lc($1) eq 'pkt2+' ) { $config{'mailout_pkt'} = 22; }
      }
      $config{'mailout'} = parse_quote2($s, '"');
    }
    # EchoOut <area>|* <path> [<msg|pkt|pkt[2[+|o]]>]
    if ($cmd eq 'echoout') {
      /^\S+\s+(\S+)\s+(.*)$/; 
      my $a = $1; my $s = $2; my %echo = ('type'=>'msg');
      $a =~ s/[*]/.*/;                         # '*' -> '.*'
      $a =~ s/[?]/./;                          # '?' -> '.'
      $s =~ s/\s+(msg|pkt(?:2(?:[o+])?)?)$//i;
      if ( defined $1 ) { 
        if ( lc($1) eq 'pkt' || lc($1) eq 'pkt2' ) { $echo{'type'} = 'pkt'; $echo{'pkt'} = 21; }
        elsif ( lc($1) eq 'pkt2o' ) { $echo{'type'} = 'pkt'; $echo{'pkt'} = 20; }
        elsif ( lc($1) eq 'pkt2+' ) { $echo{'type'} = 'pkt'; $echo{'pkt'} = 22; }
      }
      $echo{'path'} = parse_quote2($s, '"');
      $config{'echoout'}{$a} = \%echo;
    }
    # BadTics <path>
    if ($cmd eq 'badtics') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # DupeTics <path>
    if ($cmd eq 'dupetics') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # ProcTics <path>
    if ($cmd eq 'proctics') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # Queue <path>
    if ($cmd eq 'queue') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # LogFile  <file>  - log file name
    if ($cmd eq 'logfile') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # DupeFile <file>  - file to store dupebase
    if ($cmd eq 'dupefile') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # TrafFile <file>  - file to store traffic report
    if ($cmd eq 'traffile') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # RepFile  <file>  - file to store daily report
    if ($cmd eq 'repfile') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # LockFile  <file> <expire> - lock file and its expiration time in hours
    if ($cmd eq 'lockfile') { 
      if (!/^\S+\s+(?:"(.*?)(?<!\\)"|(\S+))\s+(-?[0-9.]+)$/) { cfg_error("Syntax error"); }
      $config{$cmd} = parse_quote2($1, '"').$2;
      $config{'locktime'} = $3 + 0;
    }
    # FilesBBS <name>  - name of per-area file-list
    if ($cmd eq 'filesbbs') { /^\S+\s+(\S+)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # MagicFile <name> [<format>] - freq magic list, can be multiple
    if ($cmd eq 'magicfile') {
      if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(.*?)$/) { cfg_error("Syntax error"); }
      my $mn = parse_quote($1, '"').$2; my $mf = $3;
      if (defined $mf) { 
        if ($mf !~ /^\s+"(.*?)(?<!\\)"\s*$/) { cfg_error("Expected regexp in double quotes after 'format'"); }
        $mf = $1;
      } 
      else { $mf = '^\s*@?()\s+(\S+)'; }
      push @{$config{$cmd}}, [$mn, $mf];
    }
    # MagicPass replace|update|keep|strip
    if ($cmd eq 'magicpass') {
      if (!/\S+\s+(replace|update|keep|strip)\s*$/i) { cfg_error("Syntax error"); }
      $config{$cmd} = lc $1;
    }
    # FEchoList <name>|@AreaLists@ [<format>] - fileechos list, can be multiple
    if ($cmd eq 'fecholist') {
      if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(.*?)$/) { cfg_error("Syntax error"); }
      my $fn = parse_quote($1, '"').$2; my $ff = $3;
      if (defined $ff && $ff ne '') { 
        if ($ff !~ /^\s+"(.*?)(?<!\\)"\s*$/) { cfg_error("Expected regexp in double quotes"); }
        $ff = $1;
      } 
      else { $ff = '^\s*()\s+(.*?)\s*$'; }
      push @{$config{$cmd}}, [$fn, $ff];
    }
    # Crc32  <file> [<options>] - program to calc CRC-32
    if ($cmd eq 'crc32') { 
      /^\S+\s+(.*)$/; 
      if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(?:\s+(.*?))?$/) { cfg_error("Syntax error"); }
      my $s = parse_quote($1, '"').$2;
      $config{$cmd} = $s unless ($s eq '*' || $s eq '-' || $s eq '');
      if (defined $3) {
        $s = $3;
        if ($s =~ /\bcache\b/i) { $config{'crc_cache'} = 1; $s =~ s/cache//ig; }
        if ($s =~ /\bseek=(never|always|missing)\b/i) { $config{'crc_seek'} = lc $1; $s =~ s/seek=$1//ig; }
        if ($s !~ /^\s*$/) { cfg_error("Syntax error, can't parse '$s'"); }
      }
    }
    # Group <name> <description>
    if (!$in_area && ($cmd eq 'group')) {
      if (!/^\S+\s+([A-Z]{1})\s+(.*)$/) { cfg_error("Syntax error"); }
      $config{'group'}{$1} = parse_quote2($2, '"');
    }
    # LogLevel  <level>  - log verbosity level (0..5)
    if ($cmd eq 'loglevel') { 
      if (!/^\S+\s+(\d+).*$/) { cfg_error("Number expected"); } 
      $config{$cmd} = $1; 
    }
    # ConLevel  <level>  - console verbosity level (0..5)
    if ($cmd eq 'conlevel') { 
      if (!/^\S+\s+(\d+).*$/) { cfg_error("Number expected"); } 
      $config{$cmd} = $1; 
    }
    # FileKeepDays <number>
    if ($cmd eq 'filekeepdays') { 
      if (!/^\S+\s+(\d+)/) { cfg_error("Number expected"); }
      $config{$cmd} = $1; 
    }
    # DupeKeepDays <number>
    if ($cmd eq 'dupekeepdays') { 
      if (!/^\S+\s+(\d+)/) { cfg_error("Number expected"); }
      $config{$cmd} = $1; 
    }
    # Sysop <name> - sysop name
    if (!$in_node && ($cmd eq 'sysop')) { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # Robot <string>  - robot name in requests
    if (!$in_node && ($cmd eq 'robot')) { /^\S+\s+(.*)$/; push @{$config{$cmd}}, parse_quote2($1, '"'); }
    # Origin <string>
    if ($cmd eq 'origin') { /^\S+\s+(.*)$/; $config{$cmd} = parse_quote2($1, '"'); }
    # SplitMsg <mail> [<echo>]
    if ($cmd eq 'splitmsg') {
      if (!/^\S+\s+(\d+|-)(?:\s+(\d+|-))?\s*$/) { cfg_error("Syntax error"); }
      $config{'split_mail'} = ($1 > 0) ? $1 : undef;
      $config{'split_echo'} = ($2 > 0) ? $2 : (defined $2 ? undef : $config{'split_mail'});
    }
    # FeedUnlinks keep|remove silent|warn|public
    if ( $cmd eq 'feedunlinks' ) {
      if (!/^\S+\s+(keep|remove)\s+(silent|warn|public)\s*$/i) { cfg_error("Syntax error"); }
      $config{$cmd} = lc($1)." ".lc($2);
    }
    # LastUnlinks keep|remove silent|warn|public [strict]
    if ( $cmd eq 'lastunlinks' ) {
      if (!/^\S+\s+(keep|remove)\s+(silent|warn|public)(?:\s+(strict))?\s*$/i) { cfg_error("Syntax error"); }
      $config{$cmd} = lc($1)." ".lc($2)." ".lc($3);
    }
    # GuessLongDesc on|yes|1
    if ( $cmd eq 'guesslongdesc' ) {
      $config{$cmd} = /^\S+\s+(on|yes|1)\s*$/i;
    }
    # Include <file>
    if ($cmd eq 'include') {
      if (/^\S+\s+(\S+)/) { read_config( $BASE_PATH.'/'.parse_quote2($1, '"') ); }
    }
    # Node <address> [<tic password> [<pkt password>]]
    if ( $cmd eq 'node' ) {
      if (!/^\S+\s+(\S+)(.*)/) { cfg_error("Syntax error\n"); }
      my @addr = s2addr($1);
      my $tic_pwd = '', $pkt_pwd = '';
      if (!defined $addr[0]) { cfg_error("Bad address $1\n"); }
      if ($2 =~ /^\s+(\S+)(.*)/) {
        $tic_pwd = $pkt_pwd = $1 unless $1 eq '-';
        if ($2 =~ /^\s+(\S+)/) { $pkt_pwd = $1 unless $1 eq '-'; }
      }
      $cnode = addr4s(@addr);
      %{$nodes{$cnode}} = ();
      $nodes{$cnode}{'addr'} = \@addr;
      $nodes{$cnode}{'tic_pwd'} = $tic_pwd unless $tic_pwd eq ''; 
      $nodes{$cnode}{'pkt_pwd'} = $pkt_pwd unless $pkt_pwd eq '';
      $nodes{$cnode}{'robot_pwd'} = $tic_pwd unless $tic_pwd eq ''; 
      $nodes{$cnode}{'flag_att'} = 'f';
      $nodes{$cnode}{'flag_mail'} = 'o';
      $nodes{$cnode}{'use_aka'} = use_aka(@addr);
      #while ((my $k, my $v) = each %{$config{'aka'}}) {
      #  if ( acmp(\@addr, $v) ) { my @a = s2addr($k); $nodes{$cnode}{'use_aka'} = \@a; }
      #}
      #if ( !defined $nodes{$cnode}{'use_aka'} ) {
      #  my @a = @{$config{'address'}}; $nodes{$cnode}{'use_aka'} = \@a;
      #}
      $in_node = 1; $in_area = 0;
    }
    # --- node-related commands
    if ($in_node) {
      # UseAka <aka>
      if ($cmd eq 'useaka') {
        /^\S+\s+(\S+)/;
        my @addr = s2addr($1);
        if (!defined $addr[0]) { cfg_error("Bad address $1\n"); }
        if (!my_aka(@addr)) { cfg_error("Undefined AKA $1\n"); }
        $nodes{$cnode}{'use_aka'} = \@addr;
      }
      # Sysop <name> - sysop name
      if ($cmd eq 'sysop') { /^\S+\s+(.*)$/; $nodes{$cnode}{$cmd} = parse_quote2($1, '"'); }
      # Robot  <name>|- [<robot password>]
      if ($cmd eq 'robot') {
        /^\S+\s+(?:"(.*?)(?<!\\)"|(\S+))(.*)/;
        $nodes{$cnode}{'robot'} = parse_quote($1, '"').$2 unless ($2 eq '-');
        if ($3 =~ /\s+(\S+)/) { $nodes{$cnode}{'robot_pwd'} = $1; }
      }
      # Box  <path>|-|T-Mail|T-MailLong|TheBrake!
      if ($cmd eq 'box') {
        /^\S+\s+(?:"(.*?)(?<!\\)"|(\S+))/;
        my $path = parse_quote($1, '"').$2; my $spc = $2;
        if ($spc eq '-')                { $nodes{$cnode}{$cmd} = -1; }
        elsif (lc $spc eq 't-mail')     { $nodes{$cnode}{$cmd} = 1; }
        elsif (lc $spc eq 't-maillong') { $nodes{$cnode}{$cmd} = 2; }
        elsif (lc $spc eq 'thebrake!')  { $nodes{$cnode}{$cmd} = 3; }
        else { $nodes{$cnode}{$cmd} = $path; }
      }
      # Groups <allowed list> [<auto-create group>]
      if ($cmd eq 'groups') {
        /^\S+\s+(\S+)(.*)/;
        my ($grps, $aagrp);
        $nodes{$cnode}{'groups'} = $grps = $1;
        if ($2 =~ /\s+(\S+)/) { 
          $nodes{$cnode}{'auto_add'} = $aagrp = $1; 
          if ($grps !~ /$aagrp/) { cfg_error("Node hasn't access to auto_add group"); }
        }
      }
      # Flag <attaches> [<mail>] : (h)old,(d)irect,(c)rash,(i)mm
      if ($cmd eq 'flag') {
        /^\S+\s+(\S+)(.*)/;
        my $s = $2;
        my $fl = parse_flag(lc $1);
        if (!defined $fl) { cfg_error("Bad flag '$1'\n"); } 
          else { $nodes{$cnode}{'flag_att'} = $fl; }
        if ($s =~ /\s+(\S+)/) { 
          $fl = parse_flag($1);
          if (!defined $fl) { cfg_error("Bad flag '$1'\n"); } 
            else { $nodes{$cnode}{'flag_mail'} = $fl; }
        }
      }
      # Options <list> : pause, notify, news, msg, forwreqs
      if ($cmd eq 'options') {
        my @opts = /\w+/g;
        shift @opts;
        foreach $s (@opts) {
          if ($s !~ /^(pause|notify|news|msg|forwreqs)$/i) { cfg_error("Unknown option '$s'\n"); }
            else { $nodes{$cnode}{'options'} .= "$s "; }
        }
      }
      # AreaList <file> <prio> [off] [hidden] [uncond] [format "<regexp>"]
      if ($cmd eq 'arealist') {
        if (!/\S+\s+(?:"(.*?)(?<!\\)"|(\S+))\s+(\d+)(.*?)?$/) { cfg_error("Syntax error"); }
        my $al = parse_quote($1, '"').$2; my $prio = $3; my $rest = $4;
        $nodes{$cnode}{$cmd} = {'file'=>$al, 'prio'=>$prio};
        if ($rest =~ /(off)/i) { $nodes{$cnode}{$cmd}{lc $1} = 1; }
        if ($rest =~ /(hidden)/i) { $nodes{$cnode}{$cmd}{lc $1} = 1; }
        if ($rest =~ /(uncond)/i) { $nodes{$cnode}{$cmd}{lc $1} = 1; }
        if ($rest =~ /(format)\s+"(.*?)(?<!\\)"/) { 
          if (!defined $2) { cfg_error("Expected regexp in double quotes after 'format'"); }
          $nodes{$cnode}{$cmd}{lc $1} = $2; #parse_quote($2, '"'); 
        }
        $::FRQ_ENABLED = 1;
      }
      # AutoConnect
      if ($cmd eq 'autoconnect') {
        my @tags = split /\s/; shift @tags;
        foreach my $tag (@tags) { 
          $nodes{$cnode}{$cmd}{uc $tag} = 1;
          $forward_reqs{uc $tag} = 1;
        }
      }
    } # --- /node-related commands
    # AREA <tagname> <aka> [<path> <mode>]
    if ( /^AREA\s/i ) {
      my ($tag, $aka, $path, $mode);
      if ( /^AREA\s+(\S+)\s+(\S+)\s+(?:(?:"(.*?)(?<!\\)"|(\S+))\s+(\S+))?$/i ) { 
        $tag = $1; $aka = $2; $path = parse_quote($3, '"').$4; $mode = $5;
        #$path =~ s!\\"!"!g;
      }
      elsif ( /^AREA\s+(\S+)\s+(\S+)$/i ) {
        $tag = $1; $aka = $2; undef $path; undef $mode;
      }
      else { cfg_error("Syntax error\n"); }
      my @aka = s2addr($aka);
      if ($aka[0] == undef) { cfg_error("Bad address $aka\n"); }
      # init new area
      $config::afile = $file;
      $ctag = uc $tag;
      %{$areas{$ctag}} = ();
      $areas{$ctag}{'aka'} = \@aka;
      $areas{$ctag}{'path'} = $path unless !defined $path;
      $areas{$ctag}{'mode'} = oct $mode unless !defined $mode;
      $areas{$ctag}{'dcheck'} = 'name crc';
      $in_area = 1;
    }
    # --- area-related commands
    if ($in_area) {
      # LINKS [<>*]<node1> [<>*]<node2> ...
      if ( /^LINKS\s/i ) {
        my @links = split /\s/;
        for ($j = 1; $j < @links; $j++) {
          $links[$j] =~ /^([<>*]*)(.*)$/;
          my $mode = $1;
          my @addr = s2addr($2);
          if ($addr[0] == undef) { cfg_error("Bad address $2\n"); }
          #$areas{$ctag}{'links'}{ addr2s(@addr) } = $mode;
          push @{$areas{$ctag}{'links'}}, {'addr'=>\@addr, 'mode'=>$mode};
        }
      }
      # DESCRIPTION <desc>
      elsif ( /^DESCRIPTION\s+(.+)$/i ) { $areas{$ctag}{'descr'} = parse_quote2($1, '"'); }
      # GROUP <group>
      elsif ( /^GROUP\s+(\S+)$/i ) { $areas{$ctag}{'group'} = $1; }
      # not used: EXCLUDE <regexp>
      #elsif ( /^EXCLUDE\s+(.+)$/i ) { $areas{$ctag}{'exclude'} = $1; }
      # RESTRICT <|>
      elsif ( /^RESTRICT\s+(.+)$/i ) { 
        $s = $1; 
        if ( $s !~ /^(<|>)$/ ) { cfg_error("'<' or '>' expected\n"); }
        else { $areas{$ctag}{'restrict'} = $s; }
      }
      # DUPECHECK [crc] [name] [echo]
      elsif ( /^DUPECHECK\s+(.+)$/i ) { 
        my @dc = split /\s/, $1;
        $areas{$ctag}{'dcheck'} = '';
        foreach $dc (@dc) {
          if ( $dc =~ /^(crc|name|echo)$/i ) { $areas{$ctag}{'dcheck'} .= $dc." "; }
          else { cfg_error("Unknown parameter to DupeCheck - '$dc'\n"); }
        }
      }
      # NOCHECK [crc] [size]
      elsif ( /^NOCHECK\s+(.+)$/i ) { 
        my @nc = split /\s/, $1;
        foreach $nc (@nc) {
          if ( $nc =~ /^(crc|size)$/i ) { $areas{$ctag}{'nocheck'} .= $nc." "; }
          else { cfg_error("Unknown parameter to NoCheck - '$nc'\n"); }
        }
      }
      # Magic allow|deny "<regexp>"
      # File  allow|deny "<regexp>"
      elsif (($cmd eq 'magic') || ($cmd eq 'file')) {
        if (!/^\S+\s+(allow|deny)\s+"(.*?)(?<!\\)"/) { cfg_error("Syntax error"); }
        push @{$areas{$ctag}{$cmd}}, [$1, $2];
      }
      # ONRECV <command>
      elsif ( /^ONRECV\s+(.+)$/i ) { $areas{$ctag}{'onrecv'} = parse_quote2($1, '"'); }
      # ONHATCH <command>
      elsif ( /^ONHATCH\s+(.+)$/i ) { $areas{$ctag}{'onhatch'} = parse_quote2($1, '"'); }
      # ONCREATE <command>
      elsif ( /^ONCREATE\s+(.+)$/i ) { $areas{$ctag}{'oncreate'} = parse_quote2($1, '"'); }
    } # --- /area-related commands
    # add command to list of already present commands
    if ( $type & 2 ) {
      if ($in_node) { $was_n{$cmd} = $i; } elsif ($in_area) { $was_a{$cmd} = $i; } else { $was{$cmd} = $i; }
    }
  }
  return 1;
}

sub check_config
{
  # check mandatory keywords
  while ((my $k, my $v) = each %keywords) {
    if (($v & 1) && !defined($config{$k})) { print STDERR "Fatal: Mandatory keyword '$k' not defined\n"; exit 1; }
  }
  # check pathes and files
  while ((my $k, my $v) = each %config) {
    if (ref $v eq 'ARRAY') {
      for (my $i = 0; $i < @$v; $i++) {
        if (($keywords{$k} & 0x20) && !file_name_is_absolute($v->[$i])) {
          $v->[$i] = rel2abs($v->[$i], $BASE_PATH);
        }
        if (($keywords{$k} & 0x10) && !(-d $v->[$i])) { print STDERR "Fatal: Path '$v->[$i]' for keyword '$k' not found\n"; exit 1; }
      }
    } 
    elsif (ref $v eq 'HASH') {
      foreach my $i (keys %$v) {
        if (($keywords{$k} & 0x20) && !file_name_is_absolute($v->{$i})) {
          $v->{$i} = rel2abs($v->{$i}, $BASE_PATH);
        }
        if (($keywords{$k} & 0x10) && !(-d $v->{$i})) { print STDERR "Fatal: Path '$v->{$i}' for keyword '$k' not found\n"; exit 1; }
      }
    }
    else {
      if (($keywords{$k} & 0x20) && !file_name_is_absolute($v)) {
        $v = $config{$k} = rel2abs($v, $BASE_PATH);
      }
      if (($keywords{$k} & 0x10) && !(-d $v)) { print STDERR "Fatal: Path '$v' for keyword '$k' not found\n"; exit 1; }
    }
  }
  if ( !-d $config{'queue'}."/delayed" ) {
    if ( !mkdir $config{'queue'}."/delayed", 0775 ) { print STDERR "Fatal: Can't mkdir $config{queue}/delayed\n"; exit 1; }
  }
  # check CRC-32
  my $crc32_str = 'val khokhlov';
  my $crc32_tst = 0xae16c6ab;
  if ( defined $config{'crc32'} ) {
    if ( !-x $config{'crc32'} ) { print STDERR "Fatal: CRC-32 program $config{crc32} not found or not executable\n"; exit 1; }
  }
  else {
    my $test = eval { require crc32; 1 };
    if ( !$test ) { print STDERR "Fatal: Can't load crc32 perl module or library\n$@"; exit 1; }
  }
  if ( $crc32_tst != run_crc32($crc32_str) ) { print STDERR "Fatal: CRC-32 program not found or CRC-32 test failed\n"; exit 1; }
  # check main outbound
  {
    my $path = $config{'outbound'};
    if (!defined $path) { print STDERR "Fatal: Main outbound undefined\n"; exit 1; }
    elsif (!file_name_is_absolute($v->{$i})) { $config{'outbound'} = $path = rel2abs($path, $BASE_PATH); }
    if (!-d $path) { print STDERR "Fatal: Main outbound $path not found\n"; exit 1; }
  }
  # check links
  $symlink_exists = eval { symlink("",""); 1 };
  $link_exists = eval { link("",""); 1 };
  if ( ($config{'attachmode'} eq 'link') && ($link_exists != 1) ) {
    print STDERR "Fatal: Can't use link attach mode, link() is not available\n";
    exit 1;
  }
  if ( ($config{'attachmode'} eq 'symlink') && ($symlink_exists != 1) ) {
    print STDERR "Fatal: Can't use symlink attach mode, symlink() is not available\n";
    exit 1;
  }
  if ( ($config{'filemode'} eq 'both') && ($symlink_exists != 1) ) {
    print STDERR "Fatal: Can't use 'both' file mode, symlink() is not available\n";
    exit 1;
  }
  # check magic files
  for (my $i = 0; $i < @{$config{'magicfile'}}; $i++) {
    my $v = $config{'magicfile'}[$i];
    if (!file_name_is_absolute($v->[0])) { $v->[0] = rel2abs($v->[0], $BASE_PATH); }
    if ( ! -w $v->[0] ) { 
      say 3, "Warn: Can't write to MagicFile $v->[0], exluding";
      splice @{$config{'magicfile'}}, $i--, 1;
    }
  }
  # add areas to the last config if no areas found
  $config::afile = $config::cfile if !defined $config::afile;
  # check file permissions
  if ( ! -w $config::afile ) { say 3, "Warn: Can't write to file $config::afile, area auto-add impossible"; }
  if ( ! -w $config{'dupefile'} ) { say 3, "Warn: Can't write to dupes db $config{dupefile}"; }
  if ( ! -w $config{'traffile'} ) { say 3, "Warn: Can't write to traffic db $config{traffile}"; }
  if ( ! -w $config{'repfile'} ) { say 3, "Warn: Can't write to reports db $config{repfile}"; }
  # check lock file
  if ( defined $config{'lockfile'} ) {
    if ( -e $config{'lockfile'} ) {
      if ( $config{'locktime'} < 0 || (-M $config{'lockfile'})*24 < $config{'locktime'} ) {
        say 2, "Lock file $config{lockfile} is fresh"; exit 0;
      }
      else { unlink $config{'lockfile'} or say 3, "Warn: Can't delete old lock file $config{lockfile}"; }
    }
    # create lock
    elsif (open LCK, ">$config{lockfile}") { print LCK "$$\n"; close LCK; }
  }
  # update some vars
  if ( @{$config{'robot'}} ) { $ROBOT = $config{'robot'}->[0]; }
  $SYSOP = $config{'sysop'} if ( defined $config{'sysop'} );
}

# --------------------------------------------------------------------
# function modify_node(addr \@addr, string $action)
#    updates node @addr options in memory and add config updates
#    $action ::= <+|-><option>, <option> ::= pause|notify|news|msg
sub modify_node
{
  my @addr = @{$_[0]}; my $act = $_[1];
  say 8, "modify_link(@addr, '$act')";
  # check action
  if ( $act !~ /^(\+|-)(pause|notify|news|msg|forwreqs)$/i ) { return; }
  my $act_t = $1; my $act_n = $2; my $s;
  my $s = $nodes{addr4s(@addr)}{'options'}; 
  $act_nn = "$act_n\\W+";
  # add option to end of line
  if ($act_t eq '+') { $s .= $act_n.' ' unless $s =~ /$act_n/i; }
  # delete options
  elsif ($act_t eq '-') { $s =~ s/$act_nn//gi; }
  # replace options in memory
  $s =~ s/^\s+//;
  $nodes{addr4s(@addr)}{'options'} = $s;
  # add to updates list
  if ($s ne '') { 
    $updates{addr4s(@addr)}{'upd'}{'options'} = "Options $s";
    delete $updates{addr4s(@addr)}{'del'}{'options'}; 
  }
  # no options for node
  else { 
    $updates{addr4s(@addr)}{'del'}{'options'} = 1;
    delete $updates{addr4s(@addr)}{'upd'}{'options'}; 
  }
}
# --------------------------------------------------------------------
# function modify_link(string $area, string $action)
#    updates links for area $area in memory and add config updates
#    $action ::= <+|-><link>, <link> ::= [<|>|*]<addr>
sub modify_link
{
  my $area = my $tag = $_[0]; my $act = $_[1];
  say 8, "modify_link('$tag', '$act')";
  # check action
  if ( $act !~ /^(\+|-)([<>*]*)(.*)$/i ) { return; }
  my @addr = s2addr($3);
  if ( !defined $addr[0] ) { return; }
  my $act_t = $1; my $act_f = $2; my $act_n = $3; my $s;
  # pass links array, do actions
  my $done = 0;
  my @anew = ();
    foreach $l (@{$areas{$tag}{'links'}}) {
      if ( acmp(\@addr, $l->{'addr'}) ) {
        $done = 1;
        # simply delete
        if ( $act_t eq '-' ) { 
          if ( $l->{'mode'} =~ /[*]/ ) { unfeed_area($tag, $l->{'addr'}); }
          next; 
        }
        # modify permissions
        elsif ( $act_t eq '+' ) {
          my $s2 = $l->{'mode'};
          $s2 =~ s/(<|>)//g; 
          $act_f =~ s/[*]//g;
          $s2 = $act_f.$s2;
          push @anew, {'addr'=>\@addr, 'mode'=>$s2};
        }
      }
      else { push @anew, $l; }
    }
    if ( !$done ) { push @anew, {'addr'=>\@addr, 'mode'=>$act_f}; }
  $areas{$tag}{'links'} = \@anew;
  # add to updates list
  my $s = make_links($tag);
  if ($s ne '') {
    $updates{$tag}{'upd'}{'links'} = "Links $s";
    delete $updates{$tag}{'del'}{'links'};
  }
  # no links for area
  else {
    $updates{$tag}{'del'}{'links'} = 1;
    delete $updates{$tag}{'upd'}{'links'};
  }
  # last link unlinks or only one link remains - for pass-through
  if ((@anew <= 1) && !defined $areas{$tag}{'path'}) { empty_area($tag, !@anew); }
}
# ====================================================================
sub update_add
{
  # do adds
  if ( @update_adds > 0 ) {
    open F, '>>'.$config::afile;
    foreach $tag (@update_adds) {
      say 5, "Adding newly created area '$tag' to file '$config::afile'";
      print F "# auto-added on ".POSIX::strftime("%c", localtime)."\n";
      print F "Area $tag ".addr2s(@{$areas{$tag}{'aka'}});
      if ( defined $areas{$tag}{'path'} ) { 
        my $s = $areas{$tag}{'path'};
        if ( $s =~ /\s/ ) { $s =~ s!"!\\"!; $s = '"'.$s.'"'; };
        print F " $s ".sprintf("%04o", $areas{$tag}{mode} % 0777);
      }
      print F "\n";
      if ( @{$areas{$tag}{'links'}} > 0 ) {
        print F "Links ".make_links($tag)."\n";
      }
      print F "Description $areas{$tag}{descr}\n" if defined $areas{$tag}{'descr'};
      print F "Group $areas{$tag}{group}\n" if defined $areas{$tag}{'group'};
      print F "Exclude $areas{$tag}{exclude}\n" if defined $areas{$tag}{'exclude'};
      print F "Restrict $areas{$tag}{restrict}\n"  if defined $areas{$tag}{'restrict'};
      print F "DupeCheck $areas{$tag}{dcheck}\n"  if defined $areas{$tag}{'dcheck'};
      print F "NoCheck $areas{$tag}{nocheck}\n"  if defined $areas{$tag}{'nocheck'};
      print F "OnRecv $areas{$tag}{onrecv}\n"  if defined $areas{$tag}{'onrecv'};
      print F "OnHatch $areas{$tag}{onhatch}\n"  if defined $areas{$tag}{'onhatch'};
      #print F "\n";
    }
    close F;
  }
}
# --------------------------------------------------------------------
# function update_line(string $cmd, string $object)
#    return valid config line with update for $cmd keyword of $object
sub update_line
{
  my $res;
  my ($cmd, $obj) = ($_[0], $_[1]);
  my %h = %{$updates{$obj}};
  while ( my ($act, $v) = each %h ) {
    my %hh = %$v;
    while ( my ($kv, $vv) = each %hh ) {
      chomp $vv;
      if (!defined $cmd) {
        if (($act eq 'add') || ($act eq 'upd')) { $res .= "$vv\n"; }
        next;
      }
      elsif (($act eq 'del') && ($kv eq '*')) { return ''; }
      elsif ($cmd ne $kv) { next; }
      if ($act eq 'del') { return ''; }
      if ($act eq 'upd') {
        # links for area - make single line and delete all other lines
        if ($cmd eq 'links') { $updates{$obj}{'del'}{'links'} = 1; }
        delete $updates{$obj}{$act}{$kv}; 
        return "$vv\n";
      }
    }
    if (!defined $cmd) { delete $updates{$obj}{$act}; }
  }
  return $res;
}
# --------------------------------------------------------------------
# function update_cfg(string $name)
#    update config file(s), global %updates hold modification:
#    $updates{<tag>|<node>}{add|del|upd}{<keyword>} = $new_string;
sub update_cfg
{
  # params
  my $file = $_[0];
#  my %updates = (@_ > 1) ? %{$_[1]} : ();
  # private vars
  my ($cnode, $carea);
  my $need_save = 0;
  my $I = 0; my $i = 0; my @b = ();
  # check updates records and delete errors
  while ( (my $k, my $v) = each %updates ) {
    #if (!$nodes{$k} && !$areas{$k}) { undef $updates{$k}; next; }
    while ( (my $kk, my $vv) = each %$v ) {
      undef $updates{$k}{$kk} unless ($kk eq 'add') || ($kk eq 'del') || ($kk eq 'upd');
    }
    if (keys(%{$updates{$k}}) == 0) { undef $updates{$k}; }
  }
  # exit if empty updates
  if ( keys(%updates) == 0 ) { return; }
  # read file
  open F, $file;
  my @a = <F>;
  close F;
  # for each line
  for ($i = 1; $i <= @a; $i++) {
    push @b, $a[$i-1];
    my $cmd; 
    $_ = $a[$i-1]; chomp $_;
    # strip whitespaces
    s/^\s+//; s/\s+$//;
    # ignore comments
    if ( m!^(#|;|//)! ) { next; }
    # ignore empty lines
    if ( m!^\s*$! ) { next; }
    # get keyword as first word
    ($cmd) = /^(\S+)/; $cmd = lc $cmd;
    # if were in node- or area-section and got outside keyword 
    # ...then assure adding of 'add' and 'upd' updates
    if ( (defined $cnode && !defined $nodekeys{$cmd}) ||
         (defined $carea && !defined $areakeys{$cmd}) ) {
      #pop @b;
      my $s = update_line(undef, $cnode.$carea);
      if (defined $s) { $need_save = 1; splice @b, $I, 0, $s; }
      $cnode = $carea = undef;
      #push @b, $a[$i-1];
    }
    # parse included file
    if ($cmd eq 'include') {
      if (/^\S+\s+(\S+)/) { update_cfg($1, \%updates); }
    }
    # check if got node record
    if ($cmd eq 'node') {
      $cnode = $carea = undef;
      if (/^\S+\s+(\S+)/) {
        my @addr = s2addr($1);
        if ( defined $addr[0] ) { 
          $cnode = addr4s(@addr); 
          undef $cnode unless $updates{$cnode};
        }
      }
    }
    # check if got area record
    if ($cmd eq 'area') {
      $cnode = $carea = undef;
      if (/^\S+\s+(\S+)/) { $carea = uc $1 if $updates{uc $1}; }
    }
    # check if need to update current keyword
    if (defined $cnode || defined $carea) {
      my $s = update_line($cmd, $cnode.$carea);
      if (defined $s) { $need_save = 1; pop @b; push @b, $s unless $s eq ''; }
    }
    $I = @b; # last non-empty line in _output_ file
  }
  # if were in node- or area-section and got outside file
  # ...then assure adding of 'add' and 'upd' updates
  if ( (defined $cnode) || (defined $carea) ) {
    #pop @b;
    my $s = update_line(undef, $cnode.$carea);
    if (defined $s) { $need_save = 1; splice @b, $I, 0, $s; }
    #push @b, $a[$i-1];
  }
  # save file if needed
  if ($need_save) {
    open F, ">$file";
    print F join('', @b);
    close F;
  }
}
# ====================================================================

sub write_config
{
  print "address: ".addr2s(@{$config{'address'}})."\n";
  while ( ($k, $v) = each %{$config{'aka'}} ) { print "aka $k, mask ".addr2s(@$v)."\n"; }
  foreach (@{$config{'inbound'}}) { print "inbound $_\n"; }
  print "outbound $config{outbound}\n";
  print "mail_in $config{mailin}\n";
  print "mail_out $config{mailout}, type ".(($config{'mailout_pkt'})?'pkt':'msg')."\n";
  while ( ($k, $v) = each %{$config{'group'}}) { print "group $k ($v)\n"; }
}
sub write_nodes
{
  while ( ($k, $vv) = each %nodes) {
    %v = %$vv;
    print "\nnode $k $v{tic_pwd} $v{pkt_pwd}\n";
    print "useaka ".addr2s(@{$v{'use_aka'}})."\n" if defined $v{'use_aka'};
    print "robot $v{robot} $v{robot_pwd}\n" if defined $v{'robot'};
    print "groups $v{groups} $v{auto_add}\n" if defined $v{'groups'};
    print "flag $v{flag_att} $v{flag_mail}\n" if defined $v{'flag_att'};
    print "options $v{options}\n" if defined $v{'options'};
  }
}
sub write_areas
{
  foreach $tag (keys %areas) {
    print "Area $tag ".addr2s(@{$areas{$tag}{'aka'}});
    if ( defined $areas{$tag}{'path'} ) { 
      my $s = $areas{$tag}{'path'};
      if ( $s =~ /\s/ ) { $s =~ s!"!\\"!; $s = '"'.$s.'"'; };
      print " $s $areas{$tag}{mode}"; 
    }
    print "\n";
    if ( @{$areas{$tag}{'links'}} > 0 ) {
      print "Links";
      foreach $l (@{$areas{$tag}{'links'}}) {
        print " ".$l->{'mode'}.addr2s(@{$l->{'addr'}});
      }
      print "\n";
    }
    print "Description $areas{$tag}{descr}\n" if defined $areas{$tag}{'descr'};
    print "Group $areas{$tag}{group}\n" if defined $areas{$tag}{'group'};
    print "Exclude $areas{$tag}{exclude}\n" if defined $areas{$tag}{'exclude'};
    print "Restrict $areas{$tag}{restrict}\n"  if defined $areas{$tag}{'restrict'};
    print "DupeCheck $areas{$tag}{dcheck}\n"  if defined $areas{$tag}{'dcheck'};
    print "NoCheck $areas{$tag}{nocheck}\n"  if defined $areas{$tag}{'nocheck'};
    print "OnRecv $areas{$tag}{onrecv}\n"  if defined $areas{$tag}{'onrecv'};
    print "OnHatch $areas{$tag}{onhatch}\n"  if defined $areas{$tag}{'onhatch'};
    print "\n";
  }
}

1;
