#!/usr/bin/perl
use POSIX qw(strftime _exit);
use File::Basename qw(basename dirname fileparse);
use File::Copy qw(copy move);
use File::Spec::Functions qw(rel2abs canonpath file_name_is_absolute);
# need to include my directory into @INC at compile-time
use lib dirname($0);
# my modules
use fidoaddr;
use misc;
use tic;
use msg;
use pkt;
use mgr;
use config;

$PRG_NAME      = "ffix/$^O";
$PRG_VER       = "0.02b2";
$PRG_COPYRIGHT = "Copyright (C) 2001-2003, by val khokhlov";
# ====================================================================
# GLOBAL VARS
$CONFIG_NAME = "ffix.cfg";		# default config file
$BASE_PATH = rel2abs(dirname($0)); 	# default base directory

%config;				# parsed config file
%updates;				# per-line config updates
@update_adds;				# whole area adds

%forward_reqs;				# already requested areas
$FRQ_ENABLED = 0;			# toggled by any AreaList line

%areas;					# areas hash
%nodes;					# nodes hash

$dupes;					# flat dupes array
%traffic;				# traffic hash: $area=>{...}
$report;				# flat report array
%newareas;				# new areas hash: $area=>[...]

$ROBOT = 'ffix';			# default my robot name
$SYSOP = 'Sysop';			# default my sysop name
# --------------------------------------------------------------------
%vars = ();				# parsed command-line vars
%command_line = ('toss'=>[], 'hatch'=>[], 'mgr'=>[], 'report'=>[], 'notify'=>[],
                 'purge'=>[], 'ver'=>['-v', '--version', 'version', 'ver'],
                 'help'=>['help', '-h', '--help', '-\?'],
                 'import'=>[], 'descriptions'=>[],
                 'addr'=>['-a_', '--addr='], 'bad'=>['-b', '--bad'], 
                 'config'=>['-c_', '--config='], 'days'=>['-d_', '--days='], 
                 'echo'=>['-e_', '--echo='], 'keep'=>['-k', '--keep'],
                 'quiet'=>['-q', '--quiet'], 'scan'=>['-s_?', '--scan=?'],
                 'tic'=>['-t_', '--tic=?'], 
                 'no_crc'=>['-C', '--no-crc'], 'no_dupe'=>['-D', '--no-dupe'],
                 'no_passwd'=>['-P', '--no-passwd'], 'no_size'=>['-S', '--no-size']);
$usage = <<'END';

Usage: ffix.pl <commands> [<command-flags>] [<options>]

Commands and command-flags are:
    toss			Toss incoming .tic files
      -b, --bad				Toss bad .tic's
      -D, --no-dupe 			Don't check dupes
      -C, --no-crc  			Don't check crc
      -S, --no-size 			Don't check size
      -P, --no-passwd   		Don't check passwords

    hatch			Hatch new files in fileechos
      -s [<area>], --scan[=<area>]	Scan for .desc-files in fileecho <area>
      -t <file>|-, --tic[=<file>]	Read .tic from <file> or stdin
      -D, --no-dupe 			Don't check dupes
      -C, --no-crc  			Don't check crc
      -S, --no-size 			Don't check size

    mgr				Process AreaMgr request
      -a <addr>, --addr=<addr>		Process request for <addr> from stdin
      -P, --no-passwd   		Don't check passwords

    report			Generate new files report
      -d <n>, --days=<n>		Include files for last <n> days
      -e <area>, --echo=<area>		Put report to echomail area
      -k, --keep			Don't clear report file (useful with -d)

    notify			Send notify messages
      -a <mask>, --addr=<mask>		Send to <mask> adresses

    purge			Delete already sent and old files from queue

    import descriptions		Import missing descripitons from fecholists

    help, -h, --help, -?	Display help and exit
    version, -v, --version	Display version and exit

Global options are:
    -c <file>, --config <file>	Read config from specified file
    -q, --quiet			Suppress all console output
END
# ====================================================================
# parse command-line
$_ = parse_cmdline( join(' ', @ARGV), \%command_line, \%vars );
print "$PRG_NAME ver.$PRG_VER $PRG_COPYRIGHT\n" unless $vars{'quiet'};
if ($_ ne '') {
  print STDERR "Fatal: Unparsed command-line: '$_'\n"; exit 1;
}
# first try is to set base_path to path of specified config file
if ( defined $vars{'config'} ) { 
  ($CONFIG_NAME, $BASE_PATH) = fileparse($vars{'config'});
}
# second try is to set base_path to path of executable file
elsif ( ! -f "$BASE_PATH/$CONFIG_NAME" ) { 
  # last try is to set if to current directory
  $BASE_PATH = rel2abs('.');
}

if ($vars{'help'}) { print $usage; exit; }
if ($vars{'ver'}) { exit; }

init();

if ($vars{'purge'}) { purge(); }
if ($vars{'mgr'}) { defined ($vars{'addr'}) ? read_stdin() : scan_mail(); }
if ($vars{'toss'}) { toss(); }
if ($vars{'hatch'}) { defined ($vars{'tic'}) ? hatch_read() : hatch_scan(); }
if ($vars{'notify'}) { notify(); }
if ($vars{'report'}) { report(); }
if ($vars{'import'}) { import_descr() if $vars{'descriptions'}; }

done();

# --------------------------------------------------------------------
# Signal handler for outside signals (^C)
sub handl_sig { say 1, "Got SIG$_[0] - shutting down..."; done(); _exit(1); };
# Signal handler for Perl fatals
sub handl_die { my $s = $_[0]; chomp $s; say 1, "Fatal: $s"; done(); _exit(1); };
# --------------------------------------------------------------------
# Do init stuff
sub init
{
  # read config
  init_config();
  if (-r "$BASE_PATH/$CONFIG_NAME") { read_config("$BASE_PATH/$CONFIG_NAME"); }
  else { print STDERR "Fatal: Can't read config file $BASE_PATH/$CONFIG_NAME\n"; exit 1; }
  check_config();
  # open work files
  open LOG, ">>$config{logfile}";
  if (defined LOG) { print LOG "\n-------- $PRG_NAME $PRG_VER started at ".strftime("%a, %c", localtime).", command-line '@ARGV'\n"; }
  open F, "$config{dupefile}"; read F, $dupes, -s F; close F;
  open F, "$config{repfile}"; read F, $report, -s F; close F;
  # read traffic file
  open F, "$config{traffile}"; my @arr = <F>; close F;
  foreach (@arr) {
    chomp $_;
    my @a = split /\t/;
    $traffic{$a[0]} = {'upd'=>$a[1], 'cwf'=>$a[2], 'lwf'=>$a[3], 'cmf'=>$a[4], 'lmf'=>$a[5],
                       'cwk'=>$a[6], 'lwk'=>$a[7], 'cmk'=>$a[8], 'lmk'=>$a[9]};
  }
  while ( (my $k, $v) = each %traffic ) {
    my %h = %{$v};
  }
  # install signal handlers
  $SIG{'INT'} = $SIG{'QUIT'} = \&handl_sig;
  $SIG{__DIE__} = \&handl_die;
}
# --------------------------------------------------------------------
# Do done stuff
sub done
{
  # write dupe and report files
  open F, ">$config{dupefile}"; binmode F; print F $dupes; close F;
  open F, ">$config{repfile}"; binmode F; print F $report; close F;
  # write traffic file
  open F, ">$config{traffile}"; binmode F;
  while ( (my $k, $v) = each %traffic ) {
    my %h = %{$v};
    print F "$k\t$h{upd}\t$h{cwf}\t$h{lwf}\t$h{cmf}\t$h{lmf}\t$h{cwk}\t$h{lwk}\t$h{cmk}\t$h{lmk}\n";
  }
  close F;
  # update config files
  update_cfg("ffix.cfg");
  update_add();
  # close log file
  close LOG;
  # delete lock
  if ( defined $config{'lockfile'} ) { unlink $config{'lockfile'}; }
}
# --------------------------------------------------------------------
# send %message, %news
sub do_reports
{
  # send %news
  if ( %newareas ) {
    foreach (keys %Tpl::) { undef $Tpl::{$_}; }
    %Tpl::areas = %newareas;
    foreach my $sto (keys %nodes) {
      if ( $nodes{$sto}{'options'} !~ /news/i ) { next; }
      $Tpl::aka = $nodes{$sto}{'use_aka'};
      my $s = do "tpl/newarea.pl";
      my @addr = s2addr($sto);
      my $sysop = $nodes{$sto}{'sysop'}; $sysop = 'Sysop' unless defined $sysop;
      send_mail( { from=>$ROBOT, afrom=>$Tpl::aka,
                   to=>$sysop, ato=>\@addr,
                   subj=>'New area announcement',
                   flags=>$Msg::PVT | $Msg::LOC,
                   text=>$s
                 }
               );
    }
    # send announcement to sysop
    my $s = do "tpl/newarea.pl";
    my $sysop = $config{'sysop'}; $sysop = 'Sysop' unless defined $sysop;
    send_mail( { from=>$ROBOT, afrom=>$config{'address'},
                 to=>$sysop, ato=>$config{'address'},
                 subj=>'Auto-created areas',
                 flags=>$Msg::PVT | $Msg::LOC,
                 text=>$s
               }
             );
  }
  # send %message
  foreach (keys %Tpl::) { undef $Tpl::{$_}; }
  while (my($sto, $v) = each %sendto) {
    if ( $nodes{$sto}{'options'} !~ /msg/i ) { next; }
    $Tpl::aka = $nodes{$sto}{'use_aka'}; @Tpl::areas = (); %Tpl::files = ();
    while (my($tag, $h) = each %$v) { push @Tpl::areas, $tag; $Tpl::files{$tag} = $h; }
    my $s = do "tpl/message.pl";
    my @addr = s2addr($sto);
    my $sysop = $nodes{$sto}{'sysop'}; $sysop = 'Sysop' unless defined $sysop;
    send_mail( { from=>$ROBOT, afrom=>$Tpl::aka,
                 to=>$sysop, ato=>\@addr,
                 subj=>'Sent files announcement',
                 flags=>$Msg::PVT | $Msg::LOC,
                 text=>$s
               }
             );
  }
}
# ====================================================================
# === TOSSING
# do_delayed() - process files in <queue>/delayed dir (busy when tossing)
sub do_delayed
{
  my @dfiles = glob "$config{queue}/delayed/*.*.*.*.?lo";
  foreach (@dfiles) {
    if (!/(\d+)[.](\d+)[.](\d+)[.](\d+)[.](.)[Ll][Oo]$/) { next; }
    my ($z, $n, $f, $p, $fl) = ($1, $2, $3, $4, $5);
    my @addr = ($z, $n, $f, $p);
    my $bname = ::addr2bso(@addr[0..3]);
    my @bsy_glob = glob $PAT_BSO.pat_bso('.bsy');
    #if ( defined $bname && !-f $bname.'bsy' ) {
    if ( defined $bname && !@bsy_glob ) {
      my @bglob = glob $PAT_BSO.pat_bso($fl.'lo');
      my $bname = (@bglob) ? $bglob[0] : $bname.$fl.'lo';
      say 5, "Writing delayed file $_ to $bname";
      my $bdir = dirname($bname);
      if (!-d $bdir) { if (!mkdir $bdir, 0775) { say 5, "> can't create directory $bdir"; next; } }
      move $_, $bname;
    }
  }
}
# --------------------------------------------------------------------
# toss_dir(string $path)
sub toss_dir
{
  my $path = $_[0];
  say 2, "Scanning inbound directory $path";
  my @tics = glob "$path/*.[Tt][Ii][Cc]";
  @tics = sort { substr($a, length($a)-10, 6) <=> substr($b, length($b)-10, 6) } @tics;
  foreach my $tic (@tics) {
    say 2, "Found TIC file: $tic";
    open F, $tic;
    read F, $_, -s F;
    close F;
    if (Tic::read($_) != $Tic::OK) {
      say 2, "> error in TIC file: $Tic::err";
      say 2, "> moving bad TIC file to bad-tic directory";
      move $tic, $config{'badtics'}.'/'.basename($tic);
      move $Tic::file, $config{'badtics'}.'/'.basename($Tic::file);
    }
    else {
      say 2, "Processing TIC file from ".$Tic::keys{'from'}.", area ".(uc $Tic::keys{'area'}).", file ".($Tic::keys{'file'});
      my $res = Tic::process();
      if ($res == $Tic::BAD) {
        say 2, "> error processing TIC file: $Tic::err";
        say 2, "> moving bad TIC file to bad-tic directory";
        move $tic, $config{'badtics'}.'/'.basename($tic);
        move $Tic::file, $config{'badtics'}.'/'.basename($Tic::file);
      }
      elsif ($res == $Tic::DUPE) {
        say 2, "> moving duplicate TIC file to dupe-tic directory";
        move $tic, $config{'dupetics'}.'/'.basename($tic);
        move $Tic::file, $config{'dupetics'}.'/'.basename($Tic::file);
      }
      elsif ($res == $Tic::OK) {
        if ( defined $config{'proctics'} ) {
          say 5, "> moving successfully processed TIC file to proc-tic directory";
          move $tic, $config{'proctics'}.'/'.basename($tic);
        } 
        else {
          say 5, "> deleting successfully processed TIC file";
          unlink $tic;
        }
      }
    }
  }
}
# --------------------------------------------------------------------
# tosser main function
sub toss
{
  do_delayed();
  if (defined $vars{'bad'}) { toss_dir $config{'badtics'}; }
  foreach my $dir (@{$config{'inbound'}}) { toss_dir $dir; }
  do_reports();
}
# ====================================================================
# === REPORT, NOTIFY
# daily report function
sub report
{
  my %files;
  if ( !$vars{'echo'} ) { say 2, "Unspecified area to post report"; return; }
  # parse reports array
  foreach (split /\n/, $report) {
    my @a = split /\t/;
    if ( defined $vars{'days'} && $vars{'days'} > 0) { 
      my $start = time - ($vars{'days'}-1)*24*60*60;
      $start -= $start % (24*60*60);
      next if $a[0] < $start; 
    }
    my %h = ('file'=>$a[2], 'name'=>$a[3], 'size'=>$a[4], 'desc'=>(unpack 'u', $a[6]), 'ldesc'=>(unpack 'u', $a[7]));
    push @{$files{$a[1]}}, \%h;
  }
  # generate template
  $Tpl::aka = $config{'address'};
  @Tpl::areas = sort keys %files;
  %Tpl::files = %files;
  $Tpl::force = 0;
  do 'tpl/report.pl';
  # save report
  my %msg = ('from'=>$ROBOT, 'afrom'=>$::config{'address'}, 'to'=>'All', 'ato'=>$::config{'address'},
             'area'=>uc $vars{'echo'});
  my $ok = 0;
  for (my $i = 0; $i < @Tpl::res; $i++) {
    $msg{'subj'} = 'Daily new files report';
    $msg{'subj'} .= " [".($i+1)."/".($#Tpl::res+1)."]" if @Tpl::res > 1;
    $msg{'text'} = $Tpl::res[$i];
    $ok += send_echo(\%msg);
  }
  if (@Tpl::res > 0) {
    if ($ok == @Tpl::res) { say 2, "Posted report to area $msg{area}, $ok msgs"; }
    else { say 2, "Error posting report to area $msg{area}"; }
  } else { say 2, "Nothing to report"; }
#  Pkt::close();
  # clear report file
  $report = '' unless $vars{'keep'};
}
# --------------------------------------------------------------------
# notify function
sub notify
{
  my %tosend = ();
  # create list of notify nodes
  if ( !defined $vars{'addr'} ) {
    while ( my ($sto, $v) = each %nodes ) { $tosend{$sto} = 1 if $v->{'options'} =~ /notify/i; }
  } else {
    my @a = split /,/, $vars{'addr'};
    foreach (@a) {
      my $val = 1; if (/^!/) { $val = 0; s/^!//; }
      my @addr = s2addr $_;
      if ( !defined $addr[0] ) { say 3, "Can't parse address '$_', skipping"; next; }
      while ( my ($sto, $v) = each %nodes ) { my @b = s2addr($sto); ($val) ? $tosend{$sto} = 1 : delete $tosend{$sto} if acmp(\@addr, \@b); }
    }
  }
  # do notify
  foreach my $sto (keys %tosend) {
    say 3, "Sending notify message for $sto";
    my $v = $nodes{$sto};
    #if ( $v->{'options'} !~ /notify/i ) { next; }
    $Tpl::aka = addr2s @{$v->{'use_aka'}};
    @Tpl::addr = s2addr($sto);
    $Tpl::node = $v; 
    my $to = $v->{'sysop'}; if (!defined $to) { $to = 'Sysop'; }
    do 'tpl/notify.pl';
    for (my $i = 0; $i < @Tpl::res; $i++) {
      my %msg = ('from'=>$ROBOT, 'afrom'=>$v->{'use_aka'}, 'to'=>$to, 'ato'=>\@Tpl::addr,
                 'subj'=>"Notify message for $sto", 'flags'=>$Msg::PVT|$Msg::LOC, 'text'=>$s);
      $msg{'subj'} .= " [".($i+1)."/".($#Tpl::res+1)."]" if @Tpl::res > 1;
      $msg{'text'} = $Tpl::res[$i];
      send_mail(\%msg);
    }
    $repl{'subj'} = $subj;
  }
}
# ====================================================================
# === HATCH
# scan fileechos for new files (.desc)
sub scan_dir
{
  my %area = %{$_[0]}; my $tag = $_[1];
  if ( !-d $area{'path'} ) { return; }
  my @arr = glob "$area{path}/*.desc";
  foreach my $dsc (@arr) {
    my $f = $dsc; $f =~ s/[.]desc$//i;
    if ( !-f $f ) { say 5, "Can't find file $f to hatch for $dsc"; next; }
    say 2, "Hatching file $f into area $tag";
    # read description
    open F, "$dsc";
    my $desc = <F>; chomp $desc;
    my @ldesc = <F>; chomp @ldesc;
    close F;
    # create initial .TIC values
    #my $crc = `$::config{crc32} $f`; chomp $crc;
    my $crc = sprintf "%08x", file_crc32($f);
    @Tic::ttic = (); $Tic::file = $f; %Tic::area = %area;
    %Tic::keys = ('area'=>$tag, 'file'=>basename($f), 
                  'origin'=>addr2s(@{$config{'address'}}),
                  'crc'=>$crc, 'desc'=>$desc);
    @Tic::keys{'desc'} = [$desc] if defined $desc;
    @Tic::keys{'ldesc'} = \@ldesc if @ldesc;
    # do OnHatch action
    if ( defined $area{'onhatch'} ) {
      my ($N, $n) = Tic::fnc( basename($f) );
      my $d = (@ldesc) ? join("\n", @ldesc) : $desc;
      my $cmd = parse_symb($area{'onhatch'}, '$', {
                               'A'=>$tag, 'a'=>lc $tag,
                               'N'=>$N, 'n'=>$n,
                               'P'=>$area{'path'}, 'p'=>$area{'path'},
                               'S'=>-s $f, 'C'=>lc $crc,
                               'D'=>$desc, 'd'=>$d
                             });
      say 5, "> executing command '$cmd'";
      $res = system $cmd;
      say 5, "> execution finished, exit code=".($res>>8);
    }
    # process                                           
    my $res = Tic::process($Tic::NO_COPY | $Tic::NO_DEL | $Tic::NO_REPL);
    if ($res == $Tic::DUPE) { say 2, "> duplicate file found, hatching skipped"; }
    elsif ($res == $Tic::OK) { unlink $dsc; }
    else { say 2, "> unknown error while hatching"; }
  }
}
# --------------------------------------------------------------------
sub hatch_scan
{
  my ($cm, $cp) = (0, 0);
  while ( my ($tag, $area) = each %areas ) {
    if ( defined $vars{'scan'} && (uc $vars{'scan'} ne $tag) ) { next; }
    $cm++;
    if ( defined $area->{'path'} ) { scan_dir $area, $tag; $cp++; }
  }
  if ( defined $vars{'scan'} ) {
    if ($cm == 0) { say 2, "Unknown area $vars{scan}"; }
    elsif ($cp == 0) { say 2, "Area $vars{scan} is pass-through"; }
  }
  do_reports();
}
# --------------------------------------------------------------------
sub hatch_read
{
  # read proto-TIC
  my $s = $vars{'tic'}; 
  my @arr = ();
  if ($s ne '' && $s ne '-') {
    if (!open F, $s) { say 2, "Can't read file $s"; return; }
    @arr = <F>; close F;
  } else { @arr = <STDIN>; $| = 1; print "\n"; }
  # parse keys
  @Tic::ttic = (); undef $Tic::file; %Tic::area = ();
  Tic::read_keys(join('', @arr));
  # get area
  my $tag = uc $Tic::keys{'area'};
  if (!defined $Tic::keys{'area'}) { say 2, "Keyword AREA must be defined"; return; }
  my %area = %{$areas{$tag}};
  if (!%area) { say 2, "Area $tag not found"; return; }
  %Tic::area = %area;
  # get file name, try to find in it in cur-dir, area-dir and inbounds
  my $f = $Tic::keys{'file'};
  if (!defined $f) { say 2, "Keyword FILE must be defined"; return; }
  if ((dirname($f) eq '') && !-f $f) {
    if (defined $area{'path'} && -f "$area{path}/$f") { $f = $area{'path'}."/$f"; }
    else {
      foreach $i (@{$config{'inbound'}}) { if (-f "$i/$f") { $f = "$i/$f"; last; } }
    }
  }
  if (!-r $f) { say 2, "File $f not found or can't be read"; return; }
  # try to guess, whether file is already in that directory
  say 2, "Hatching file $f into area $Tic::keys{area}";
  my $flags = 0;
  if (defined $area{'path'}) {
    my $name = basename($f);
    if (-f "$area{path}/$name") {
      my @a_src = stat $f; my @a_trg = stat "$area{path}/$name";
      if (@a_src[0..10] == @a_trg[0..10]) {
        say 3, "> warn: file $f already exists in $area{path}, skip copying";
        $flags |= $Tic::NO_COPY | $Tic::NO_DEL; 
      }
    }
  }
  # check values, fill if required
  if (!defined $Tic::keys{'crc'}) { 
    #$_ = `$::config{crc32} $f`; chomp; 
    $Tic::keys{'crc'} = sprintf "%08x", file_crc32($f);
  }
  if (!defined $Tic::keys{'origin'}) {
    $Tic::keys{'origin'} = addr2s(@{$config{'address'}}); 
  }
    # do OnHatch action
    if ( defined $area{'onhatch'} ) {
      my ($N, $n) = Tic::fnc( basename($f) );
      my $D = join("\n", @{$keys{'desc'}});
      my $d = (@{$keys{'ldesc'}}) ? join("\n", @{$keys{'ldesc'}}) : $D;
      my $cmd = parse_symb($area{'onhatch'}, '$', {
                               'A'=>$tag, 'a'=>lc $tag,
                               'N'=>$N, 'n'=>$n,
                               'P'=>$area{'path'}, 'p'=>$area{'path'},
                               'S'=>-s $f, 'C'=>lc $Tic::keys{'crc'},
                               'D'=>$D, 'd'=>$d, 'M'=>$keys{'magic'}
                             });
      say 5, "> executing command '$cmd'";
      $res = system $cmd;
      say 5, "> execution finished, exit code=".($res>>8);
    }
  # try to do hatching
  $Tic::keys{'file'} = basename($f); $Tic::file = $f;
  my $res = Tic::process($flags);
  if ($res == $Tic::DUPE) { say 2, "> duplicate file found, hatching skipped"; }
  elsif ($res != $Tic::OK) { say 2, "> unknown error while hatching"; }
  do_reports();
}
# ====================================================================
# === PURGE
sub purge
{
  # to-do: find better stuff ;)
  my $FC = ($File::Basename::Fileparse_igncase) ? sub { lc $_[0]; } : sub { $_[0]; };
  # ;-)
  do_delayed();
  # create list of files to keep in queue
  my %keep = ();
  # create list of queue dirs
  my @dirs = glob "$config{queue}/*.*.*.*";
  # process queue
  foreach (@dirs) {
    my @addr = split /\./, basename($_);
    say 5, "Processing queue directory $_ for ".addr2s(@addr);
    # create list of .tic's
    @tics = glob "$_/*.[Tt][Ii][Cc]";
    # process .tic
    foreach my $tic (@tics) {
      my $s = '';
      open TIC, $tic; binmode TIC; read TIC, $s, -s TIC; close TIC;
      my ($name) = $s =~ /^file\s+(.*?)\r$/mi;
      # delete old .tic and its file
      if ( ($config{'filekeepdays'} > 0) && (-M $tic > $config{'filekeepdays'}) ) {
        say 3, "Deleting old TIC $tic and file $name for ".addr2s(@addr);
        # file in <queue>/z.n.f.p dir
        if ( -e "$_/$name" ) {
          if (!unlink "$_/$name") { say 2, "> can't delete file $_/$name"; }
        }
        # tic
        if (!unlink $tic) { say 2, "> can't delete file $tic"; }
      }
      else { $keep{&$FC($name)} = 1; }
    }
    # kill empty dirs
    rmdir;
  }
  # kill unused files in queue dir
  opendir DIR, $config{'queue'};
  while ($_ = readdir DIR) {
    if (!-f "$config{queue}/$_") { next; }
    if (!$keep{&$FC($_)}) { 
      say 3, "Deleting unused file $_ in queue dir";
      if (!unlink "$config{queue}/$_") { say 2, "> can't delete file $config{queue}/$_"; }
    }
  }
  closedir DIR;
  # delete old records from dupes database
  my @arr = split /\n/, $dupes;
  $dupes = '';
  my $start = time() - $config{'dupekeepdays'}*24*60*60;
  for (@arr) {
    my ($echo, $file, $crc, $time) = split /\t/;
    $dupes .= $_."\n" unless ($time < $start);
  }
}
# ====================================================================
# === IMPORT
sub import_descr
{
  while ( my ($tag, $h) = each %areas ) {
    if ($tag =~ /^[?*]/) { next; }
    my $grp = $h->{'group'};
    $grp = '' if (!$areas{'?'.$grp} || !defined $areas{'?'.$grp}{'descr'});
    if ( !defined $h->{'descr'} ||
         ($h->{'descr'} eq $areas{'?'.$grp}{'descr'}) 
       ) 
    {
      my $s = find_descr($tag);
      $updates{$tag}{'upd'}{'description'} = "Description \"$s\"" if defined $s;
    }
  }
}
